# GNU Solfege - free ear training software
# Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008  Tom Cato Amundsen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import errno
import gobject
import locale
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import webbrowser

from solfege import mpd
from solfege import soundcard

from solfege import abstract
from solfege import cfg
from solfege import dataparser
from solfege import gu
from solfege import i18n
from solfege import lessonfile
from solfege import osutils
from solfege import reportlib
from solfege import utils

import solfege

def check_rcfile():
    """See default.config for rcfileversion values, meanings and
    a description of how to add config variables.
    """
    rcfileversion = 15
    if cfg.get_int("app/rcfileversion") > rcfileversion:
        cfg.drop_user_config()
        return
    if cfg.get_int("app/rcfileversion") <= 1:
        if not "example-files" in cfg.get_string('config/lessoncollections'):
            cfg.set_string('config/lessoncollections', 
                "%s example-files" % cfg.get_string('config/lessoncollections'))
    if cfg.get_int("app/rcfileversion") <= 5:
        # This is more complicated that necessary to fix an old
        # error.
        if cfg.get_string("sound/commandline"):
            cfg.del_key("sound/commandline")
    if cfg.get_int("app/rcfileversion") <= 3:
        cfg.set_list("config/lessoncollections",
            cfg.get_string("config/lessoncollections").split())
    if cfg.get_int("app/rcfileversion") <= 4:
        cfg.del_key("config/web_browser")
    if sys.platform == 'win32':
        if cfg.get_string('sound/wav_player'):
            cfg.del_key('sound/wav_player')
    if cfg.get_int("app/rcfileversion") <= 5:
        cfg.set_string("mainwin/history_back_ak", "<alt>Left")
        cfg.set_string("mainwin/history_forward_ak", "<alt>Right")
        cfg.set_string("mainwin/history_reload_ak", "<ctrl>r")
    if cfg.get_int("app/rcfileversion") <= 6:
        cfg.set_list("config/lessoncollections", ['solfege', 'user'])
    if cfg.get_int("app/rcfileversion") <= 7:
        cfg.set_int("rhythm/countin_perc", 80)
    if cfg.get_int("app/rcfileversion") <= 8:
        cfg.del_key("singinterval/highest_tone")
        cfg.del_key("singinterval/lowest_tone")
        cfg.del_key("melodicinterval/highest_tone")
        cfg.del_key("melodicinterval/lowest_tone")
        cfg.del_key("harmonicinterval/highest_tone")
        cfg.del_key("harmonicinterval/lowest_tone")
    if cfg.get_int("app/rcfileversion") <= 9:
        cfg.del_section("mainwin")
    if cfg.get_int("app/rcfileversion") <= 10:
        cfg.del_section("lessoncollections")
        cfg.del_key("config/lessoncollections")
        for n in cfg.iterate_sections():
            cfg.del_key("%s/lessoncollection" % n)
            cfg.del_key("%s/lessonfile" % n)
    if cfg.get_int("app/rcfileversion") <= 11:
        for s in ('rhythm', 'rhythmtapping2'):
            cfg.del_key("%s/countin_perc" % s)
            cfg.del_key("%s/rhythm_perc" % s)
    if cfg.get_int("app/rcfileversion") <= 12:
        cfg.del_key("sound/card_info")
    if cfg.get_int("app/rcfileversion") <= 13:
        cfg.del_key("config/lowest_instrument_velocity")
        cfg.del_key("config/middle_instrument_velocity")
        cfg.del_key("config/highest_instrument_velocity")
        cfg.del_key("config/preferred_instrument_velocity")
    if cfg.get_int("app/rcfileversion") <= 14:
        # We have to split the midi_to_wav_cmd into two strings, and
        # moving the options to *_options, so that midi_to_wav_cmd only
        # have the name of the binary. This to allow binaries in dirs
        # with spaces.
        for k in ("midi_to_wav", "wav_to_mp3", "wav_to_ogg"):
            v = cfg.get_string("app/%s_cmd" % k).split(" ")
            cfg.set_string("app/%s_cmd" % k, v[0])
            cfg.set_string("app/%s_cmd_options" % k, " ".join(v[1:]))
    if cfg.get_int("app/rcfileversion") <= 15:
        for k in ("midi", "wav", "mp3", "ogg"):
            v = cfg.get_string("sound/%s_player" % k).split(" ")
            cfg.set_string("sound/%s_player" % k, v[0])
            cfg.set_string("sound/%s_player_options" % k,
                " ".join(v[1:]))
    cfg.set_int("app/rcfileversion", rcfileversion)
    try:
        a = mpd.notename_to_int(cfg.get_string("user/lowest_pitch"))
        b = mpd.notename_to_int(cfg.get_string("user/highest_pitch"))
    except mpd.InvalidNotenameException:
        if cfg.get_string("user/sex") == "male":
            cfg.set_string("user/highest_pitch", "e'")
            cfg.set_string("user/lowest_pitch", "c")
        else:
            cfg.set_string("user/highest_pitch", "e''")
            cfg.set_string("user/lowest_pitch", "c'")


class SolfegeApp(cfg.ConfigUtils):
    def __init__(self, options, lessonfile_manager):
        """
        options -- command line options parsed by optparse
        """
        cfg.ConfigUtils.__init__(self, 'solfege-app')
        self.m_startup_complete = False
        lessonfile.MusicBaseClass.temp_dir = tempfile.mkdtemp(prefix="solfege-")
        os.environ['SOLFEGETEMPDIR'] = lessonfile.MusicBaseClass.temp_dir
        # test_mode is when we are running a test from the Tests menu
        self.m_test_mode = False
        self.m_options = options
        self.m_teachers = {}
        self.m_running_exercise = None
        self.lessonfile_manager = lessonfile_manager
        self.m_sound_init_exception = None
        #
        self.m_userman_language = "C"
        for lang in i18n.langs():
            if os.path.isdir(os.path.join('help', lang)):
                self.m_userman_language = lang
                break
        # We have to check if lesson files have changed since last time we practised.
        self._pp = lessonfile_manager.parse(options.debug if options else None)
        gobject.idle_add(self.reparse_lessonfiles)
    def nonidle_reparse(self):
        total = len(self.lessonfile_manager.m_uiddb) * 2
        count = 0.0
        for uuid, data in self.lessonfile_manager.parse(self.m_options.debug):
            yield count / total
            count += 1
        from solfege import lessonfilegui
        lessonfilegui.handle_lesson_id_crash(lessonfile.manager)
        for uuid in self.lessonfile_manager.iterate_lesson_ids():
            data = lessonfile.manager.m_uiddb[uuid]
            solfege.db.set_lessonfile_data(uuid, data['header']['title'],
                    data['header']['module'], data['filename'])
            yield count / total
            count += 1
    def reparse_lessonfiles(self):
        """
        Parse all lesson files when we are idle.
        """
        # Check that solfege.win has been created. It is possible that
        # it is not created the first time we are idle, if the user is
        # prompted a question during MainWin construction
        if not getattr(solfege, 'win', None):
            return True
        try:
            uuid, data = self._pp.next()
            return True
        except StopIteration:
            solfege.win.handle_lesson_id_crash()
            self._pp = lessonfile.manager.iterate_lesson_ids()
            gobject.idle_add(self.check_for_changed_lessonfiles)
            return False
    def check_for_changed_lessonfiles(self):
        """
        Feed the parsed lesson file data to the database. We do
        it when we are idle.
        """
        try:
            uuid = self._pp.next()
            data = lessonfile.manager.m_uiddb[uuid]
            solfege.db.set_lessonfile_data(uuid, data['header']['title'],
                    data['header']['module'], data['filename'])
            return True
        except StopIteration:
            del self._pp
        self.m_startup_complete = True
        # If the user is really fast, or the computer not very idle, the
        # page displayed might need to be updated. It can contain warning
        # signs instead of links to some exercises.
        if solfege.win.m_viewer in ('frontpage', 'frontpage'):
            solfege.win.get_view().display_data(solfege.win.get_view()._page_tree)
        return False
    def setup_sound(self):
        if sys.platform == 'win32' and \
                    cfg.get_string("sound/type") == "sequencer-device":
            # just in case c:\home\.solfegerc is wrong
            cfg.set_string("sound/type", "winsynth")
        if self.m_options.no_sound \
           or cfg.get_string("sound/type") == "fake-synth":
            soundcard.initialise_using_fake_synth(self.m_options.verbose_sound_init)
        elif cfg.get_string("sound/type") == "winsynth":
            try:
                soundcard.initialise_winsynth(cfg.get_int("sound/synth_number"),
                      verbose_init=self.m_options.verbose_sound_init)
            except ImportError, e:
                self.display_sound_init_error_message(e)
                cfg.set_string("sound/type", "fake-synth")
                soundcard.initialise_using_fake_synth(True)
                return
            except RuntimeError, e:
                # We can get here if winmidi.output_devices() in winsynth
                # __init__ returns no devices. Don't know when, but it could
                # happen.
                solfege.win.display_exception_message(e)
                cfg.set_string("sound/type", "fake-synth")
                soundcard.initialise_using_fake_synth(True)
                return
            if cfg.get_int("sound/synth_number") != soundcard.synth.m_devnum:
                solfege.win.display_error_message2(_("MIDI setup"), _("MIDI Device %(olddev)i not available. Will use device %(newdev)i.") % {'olddev': cfg.get_int("sound/synth_number"), 'newdev': soundcard.synth.m_devnum})
                cfg.set_int("sound/synth_number", soundcard.synth.m_devnum)
        elif cfg.get_string("sound/type") == "external-midiplayer":
            soundcard.initialise_external_midiplayer(
                    cfg.get_string("sound/midi_player"),
                    verbose_init=self.m_options.verbose_sound_init)
            soundcard.synth.error_report_cb = solfege.win.display_error_message
        elif cfg.get_string("sound/type") == '':
            solfege.win.display_error_message(
                _("You should configure sound from the 'Sound' page "
                  "of the preferences window."))
        elif cfg.get_string("sound/type") == "sequencer-device":
            try:
                soundcard.initialise_devicefile(
                             cfg.get_string("sound/device_file"),
                             cfg.get_int("sound/synth_number"),
                             verbose_init=self.m_options.verbose_sound_init)
            except (soundcard.SoundInitException, OSError, ImportError), e:
                self.m_sound_init_exception = e
                soundcard.initialise_using_fake_synth(True)
        if cfg.get_string("programs/csound") == "AUTODETECT":
            for p in osutils.find_csound_executables():
                cfg.set_string("programs/csound", p)
                break
            else:
                # If not csound binary was found, then we set the string empty.
                # This means that autodetection will only happen the first time
                # you run the program. But later will newly installed binaries
                # be shown in the combo box of the preferences window.
                cfg.set_string("programs/csound", "")
        if cfg.get_string("programs/mma") == "AUTODETECT":
            for p in osutils.find_mma_executables(cfg.get_list("app/win32_ignore_drives")):
                cfg.set_string("programs/mma", p)
                break
            else:
                cfg.set_string("programs/mma", "")
    def display_sound_init_error_message(self, e):
        if isinstance(e, soundcard.SoundInitException):
            solfege.win.display_error_message(
            """%s""" % str(e).decode(locale.getpreferredencoding(), 'replace'))
        elif isinstance(e, ImportError):
            solfege.win.display_error_message2(str(e), _("You should configure sound from the preferences window, and try to use an external midi player. Or try to recompile the program and check for error messages to see why the module is not built."))
        elif e.errno == errno.EACCES:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "The errno EACCES indicates that you don't have write "
                "permission to the device."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
        elif e.errno == errno.EBUSY:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "It seems like some other program is using the device. You "
                "should try to quit that other program and restart Solfege."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
        else:
            solfege.win.display_error_message(
                "The sound init failed: %s\n"
                "You should configure sound from the 'Sound' page of "
                "the preferences window.\n\n"
                "It is also possible that the OS sound setup is incorrect."
                % str(e).decode(locale.getpreferredencoding(), 'replace'))
    def play_happy_sound(self):
        utils.play_music(r"\staff\relative c'{c16 e g a}", 180, 8,
               cfg.get_int('config/preferred_instrument_volume'))
    def play_sad_sound(self):
        utils.play_music(r"\staff\relative c'{<c,,8 cis>", 80, 58,
               cfg.get_int('config/preferred_instrument_volume'))
    def please_help_me(self):
        if isinstance(solfege.win.get_view(), abstract.Gui):
            # If the view visible is an exercise, when we check if the
            # lesson file header define a specific help page.
            if self.m_teachers[self.m_running_exercise].m_P.header.help:
                self.handle_href('%s.html' % self.m_teachers[self.m_running_exercise].m_P.header.help)
            else:
                # if not, we display help page named the same as the
                # exercise module
                self.handle_href('%s.html' % solfege.win.m_viewer)
    def show_exercise_theory(self):
        if self.m_teachers[self.m_running_exercise].m_P.header.theory:
            solfege.win.display_docfile("%s.html" % self.m_teachers[self.m_running_exercise].m_P.header.theory, "")
    def _practise_lesson_id(self, lesson_id, urlobj=None):
        """
        return the module name.
        """
        module = solfege.db.get_field(lesson_id, 'module')
        if self.m_running_exercise:
            solfege.win.box_dict[self.m_running_exercise].on_end_practise()
        if module not in self.m_teachers:
            self.create_teacher(module)
        if module not in solfege.win.box_dict:
            solfege.win.initialise_exercise(self.m_teachers[module])
        self.m_teachers[module].set_lessonfile(
            solfege.db.get_field(lesson_id, 'filename'))
        if self.m_teachers[module].m_P:
            solfege.win.activate_exercise(module, urlobj)
        self.m_running_exercise = module
        self.m_teachers[module].g_view = solfege.win.box_dict[module]
        solfege.win.show_help_on_current()
        return module
    def practise_lesson_id(self, lesson_id):
        def cleanup():
            module = solfege.db.get_field(lesson_id, 'module')
            self.m_teachers[module].m_P = None
            solfege.win.box_dict[module].practise_box.set_sensitive(False)
            solfege.win.box_dict[module].config_box.set_sensitive(False)
            solfege.win.box_dict[module].action_area.set_sensitive(False)
            solfege.win.box_dict[module].std_buttons_end_practise()
        try:
            module = self._practise_lesson_id(lesson_id)
        except (lessonfile.LessonfileParseException,
                dataparser.DataparserException), e:
            cleanup()
            solfege.win.display_exception_message(
                e, lessonfile=solfege.db.get_field(lesson_id, 'filename'))
            return
        # The try-except clause above will not catch trying to parse a
        # lesson_id that does not exist. This because
        # abstract.Teacher.parse_lessonfile will catch the exception and
        # display a message informing the user. So we have to check m_P here.
        # FIXME it can look as there is room for some redesign here...
        if not solfege.win.box_dict[module].m_t.m_P:
            cleanup()
            return
        if 'm_discards' in dir(self.m_teachers[module].m_P):
            for msg in self.m_teachers[module].m_P.m_discards:
                print >> sys.stderr, msg
        solfege.win.box_dict[module].practise_box.set_sensitive(True)
        solfege.win.box_dict[module].config_box.set_sensitive(True)
        solfege.win.box_dict[module].action_area.set_sensitive(True)
        solfege.win.box_dict[module].on_start_practise()
        w = solfege.win.g_ui_manager.get_widget("/Menubar/HelpMenu/PerExerciseHelp/HelpTheory")
        if w:
            w.set_sensitive(bool(self.m_teachers[module].m_P.header.theory))
        return module
    def test_lesson_id(self, lesson_id):
        self.m_test_mode = True
        module = self.practise_lesson_id(lesson_id)
        solfege.win.box_dict[module].on_start_practise()
        solfege.win.enter_test_mode()
    def handle_href(self, href, display_docfile_set_adj=1):
        # All generatet docbook documents are in utf-8
        urlobj = utils.Url(href)
        if not urlobj.protocol:
            if urlobj.filename.endswith(".midi"):
                if os.path.exists(os.path.join(solfege.win.box_dict['docviewer'].m_htmlwidget.m_document_wd, urlobj.filename)):
                    midifilename = urlobj.filename
                elif len(urlobj.filename.split('-')) > 1:
                    # Filename-nl.midi => Filename.midi
                    v = urlobj.filename.split('-')
                    w = v[-1].split('.')
                    midifilename = "-".join(v[:-1])+"."+w[1]
                else:
                    midifilename = urlobj.filename

                soundcard.play_mediafile("midi", os.path.join(
                    solfege.win.box_dict['docviewer'].m_htmlwidget.m_document_wd,
                    midifilename))
            else:
                solfege.win.display_docfile(urlobj.filename, urlobj.anchor)
        elif urlobj.protocol == 'solfege':
            self.handle_solfege_href(urlobj)
        elif urlobj.protocol == 'http':
            try:
                webbrowser.open_new(href)
            except Exception, e:
                solfege.win.display_error_message2(_("Error opening web browser"), str(e))
        elif urlobj.protocol == 'mailto':
            if self.get_string("config/mua"):
                try:
                    cmdline = self.get_string("config/mua") % urlobj.href
                except TypeError:
                    cmdline = "%s %s" % (self.get_string("config/mua"), urlobj.href)
                os.system(cmdline)
            else:
                gu.dialog_ok(_("You have not selected email program. You can do so in the preferences window."))
        else:
            gu.dialog_ok("unknown link type: %s" % urlobj.protocol)
    def handle_solfege_href(self, urlobj):
        """
        This method handles three kins of urls:
        solfege:practise/lesson_id
            practise the exercise with the lesson id 'lesson_id'
        solfege:practise/path/to/lessonfile
            Practise a specific lessonfile. Relative path:
                solfege:practise/lesson-files/chord-min-major
            Absolute path:
                solfege:practise//home/user/src/solfege/lesson-files/filename
        """
        solfege.win.hide_help_on_current()
        # the Url class does not know anything about lesson ids, but we
        # assume that lesson_ids don't include the '/' character.
        if '/' not in urlobj.lessonfile:
            self.practise_lesson_id(urlobj.lessonfile)
            return
        if urlobj.lessonfile:
            if not os.path.isfile(urlobj.lessonfile):
                solfege.win.display_error_message(_("Lessonfile not found:\n%s") % urlobj.lessonfile)
                return
        p = lessonfile.LessonfileCommon()
        p.parse_file(urlobj.lessonfile)
        self._practise_lesson_id(p.header.lesson_id, urlobj)
        # This loop changes variables in the lesson file header if they
        # have set values in the url.
        for key, val in urlobj.config.iteritems():
            try:
                self.m_teachers[p.header.module].m_P.header[key] = eval(val)
            except SyntaxError:
                gu.dialog_ok(_("Bad url parameter in link: '%s'.\nTrying to continue anyway.") % ("%s=%s" % (key, val)))
        solfege.win.box_dict[p.header.module].on_start_practise()
    def create_teacher(self, modulename):
        """
        Create the teacher in 'modulename' and add it to self.m_teachers.
        """
        m = __import__("solfege.exercises.%s" % modulename, fromlist=("solfege.exercises.%s" % modulename,), level=0)
        self.m_teachers[modulename] = m.Teacher(modulename)
    def reset_exercise(self, w=None):
        """
        Call on_end_practise, and then on_start_practise in
        the currently active exercise, if we have a exercise.
        """
        if isinstance(solfege.win.get_view(), abstract.Gui):
            solfege.win.get_view().on_end_practise()
            solfege.win.get_view().on_start_practise()
    def quit_program(self):
        try:
            cfg.sync()
        except IOError, e:
            solfege.win.display_exception_message(e)
        if isinstance(solfege.win.get_view(), abstract.Gui):
            g = solfege.win.get_view()
            # Check that the lesson file has a header, because if the
            # user tries to quit the program after parsing a bad lesson
            # file, we cannot call end_practise() without risking more
            # exceptions.
            if g.m_t.m_P and getattr(g.m_t.m_P, 'header', None):
                g.on_end_practise()
        solfege.db.conn.close()
        if soundcard.synth:
            soundcard.synth.close()
        shutil.rmtree(lessonfile.MusicBaseClass.temp_dir, True)
    def export_training_set(self, export_data, export_dir, output_format):
        """
        This function requires a program that can create WAV files
        from MIDI files and MP3 files from WAV.
        """
        def delay(n, tempo):
            """
            tempo is a dict of two integers
            """
            track = mpd.Track()
            track.set_bpm(*tempo)#self.get_int('config/default_bpm'))
            track.note(mpd.Rat(n, 4), 80, 0)
            soundcard.synth.play_track(track)
        track_idx = 0
        num = sum([x['count'] for x in export_data])
        # MainWin will set this to True if the user want to cancel
        # the export.
        self.m_abort_export = False
        report = reportlib.Report()
        report.append(reportlib.Heading(1, "Exported exercises"))
        table = reportlib.Table()
        report.append(table)
        for lesson_info in export_data:
            lesson_id = lesson_info['lesson_id']
            module = solfege.db.get_field(lesson_id, 'module')
            if module not in self.m_teachers:
                self.create_teacher(module)
            p = self.m_teachers[module].lessonfileclass()
            p.parse_file(solfege.db.get_field(lesson_id, 'filename'))
            for c in range(lesson_info['count']):
                if module == 'idbyname':
                    p.select_random_question()
                    if p.header.lesson_heading:
                        s = p.header.lesson_heading
                    else:
                        s = p.header.title
                    table.append_row("%i" % track_idx,
                                     p.get_question().name,
                                     s)
                    soundcard.start_export(os.path.join(
                            export_dir, "track-%i.mid" % track_idx))
                    for n in range(lesson_info.get('repeat', 1)):
                        p.play_question()
                        if n != lesson_info.get('repeat', 1) - 1:
                            if 'delay' in lesson_info:
                                delay(lesson_info['delay'], p.get_tempo())
                    soundcard.end_export()
                elif module in ('melodicinterval', 'harmonicinterval'):
                    t = self.m_teachers[module]
                    t.set_lessonfile(solfege.db.get_field(lesson_id, 'filename'))
                    t.start_practise()
                    t.new_question("c", "c''")
                    t.q_status = t.QSTATUS_SOLVED
                    try:
                        table.append_row("%i" % track_idx, "%s" % utils.int_to_intervalname(t.m_interval))
                    except AttributeError:
                        table.append_row("%i" % track_idx, "%s" % (" + ".join([utils.int_to_intervalname(q, False, True) for q in t.m_question])))
                    soundcard.start_export(os.path.join(
                            export_dir, "track-%i.mid" % track_idx))
                    for n in range(lesson_info.get('repeat', 1)):
                        t.play_question()
                        if n != lesson_info.get('repeat', 1) - 1:
                            if 'delay' in lesson_info:
                                delay(lesson_info['delay'],
                                    (self.get_int('config/default_bpm'), 4))
                    soundcard.end_export()
                else:
                    logging.warning("export_training_set:ignoring exercise with module='%s'" % module)
#####
                def do_convert(from_format, to_format):
                    """
                    Return False if we think the convert failed.
                    """
                    app_cfg_name = "app/%s_to_%s_cmd" % (from_format, to_format)
                    if from_format == 'midi':
                        from_ext = 'mid'
                    else:
                        from_ext = from_format
                    to_ext = to_format
                    if not cfg.get_string(app_cfg_name):
                        solfege.win.display_error_message2("Config variable not defined", "The missing or empty variable was '%s'" % app_cfg_name)
                        return False
                    try:
                        inout = {
                            'in': os.path.join(export_dir,
                                    "track-%i.%s" % (track_idx, from_ext)),
                            'out': os.path.join(export_dir,
                                    "track-%i.%s" % (track_idx, to_ext))}
                        opts = cfg.get_string(app_cfg_name + '_options').split(" ")
                        opts = [x % inout for x in opts]
                        # For some reasong setting the executable arg does
                        # not work for Python 2.5.2
                        try:
                            subprocess.call(
                                [cfg.get_string(app_cfg_name)] + opts)
                        except OSError, e:
                            raise osutils.BinaryForMediaConvertorException(app_cfg_name,
                                cfg.get_string(app_cfg_name), e)

                        if os.path.exists(os.path.join(export_dir, "track-%i.%s" % (track_idx, to_ext))):
                            os.remove(os.path.join(export_dir, "track-%i.%s" % (track_idx, from_ext)))
                        else:
                            # This means that the program failed to generate
                            # the WAV file. We set output_format to 'midi'
                            # because we don't want to display this error for
                            # every single file.
                            output_format = 'midi'
                            solfege.win.display_error_message2("External program must have failed", "The file in %(from)s format was not generated from the %(to)s file as expected. Please check your setup in the preferences window (CTRL-F12)." % {'to':to_format.upper(), 'from': from_format.upper()})
                    except (TypeError, KeyError):
                        solfege.win.display_error_message2("%(from)s to %(to)s config error", "There was a format string error. Will not generate WAV files. Please check the app/midi_to_wav_cmd config variable." % {'from': from_format, 'to': to_format})
                        output_format = 'midi'
                    return True
#####
                if output_format in ('mp3', 'wav', 'ogg'):
                    do_convert('midi', 'wav')
                if output_format in ('mp3', 'ogg'):
                    if not do_convert('wav', output_format):
                        output_format = 'wav'
                track_idx += 1
                yield 1.0 * track_idx / num
                if self.m_abort_export:
                    del self.m_abort_export
                    return
        reportlib.HtmlReport(report, os.path.join(export_dir, "toc.html"))
    def sheet_gen_questions(self, count, sdict):
        """
        count -- how many questions should we generate. We use this value
                 and not sdict['count'] because sometimes the app has some
                 questions, and just need a few more.
        """
        module = solfege.db.get_field(sdict['lesson_id'], 'module')
        if module not in self.m_teachers:
            self.create_teacher(module)
        p = self.m_teachers[module].lessonfileclass()
        p.parse_file(solfege.db.get_field(sdict['lesson_id'], 'filename'))
        if module == 'idbyname':
            for x in self._sheet_gen_question_idbyname(p, count, sdict):
                yield x
        else:
            assert module in ('harmonicinterval', 'melodicinterval')
            for x in self._sheet_gen_question_interval(module, p, count, sdict):
                yield x
    def _sheet_gen_question_idbyname(self, p, count, sdict):
        """
        yield count dicts, where each dict contain the data needed to
        print both the teachers and the students question.
        """
        counts = {}.fromkeys(range(len(p.m_questions)), 0)
        for x in range(count):
            while 1:
                p.select_random_question()
                if counts[p._idx] >= 1.0 * sdict['count'] / len(p.m_questions):
                    continue
                counts[p._idx] += 1
                break
            ret = {'question': {}, 'answer': {}}
            if sdict['qtype'] == 0:
                ret['question']['name'] = "...."
                ret['answer']['name'] = p.get_question().name
                ret['question']['music'] = p.get_lilypond_code()
                ret['answer']['music'] = p.get_lilypond_code()
                yield ret
            else:
                assert sdict['qtype'] == 1
                ret['question']['name'] = p.get_question().name
                ret['answer']['name'] = p.get_question().name
                ret['answer']['music'] = p.get_lilypond_code()
                ret['question']['music'] = p.get_lilypond_code_first_note()
                yield ret
    def _sheet_gen_question_interval(self, module, p, count, sdict):
        # FIXME in the idbyname we count how many times each question
        # has been selected, so get an even selection. We don't do it
        # here at the moment, because we need to descide what we really want.
        teacher = self.m_teachers[module]
        teacher.set_lessonfile(solfege.db.get_field(sdict['lesson_id'],
                                                          'filename'))
        teacher.start_practise()
        for x in range(count):
            teacher.new_question("c'", "c''")
            # quick hack to use this for both melodic and harmonic intervals
            if module == 'melodicinterval':
                teacher.m_interval = teacher.m_question[0]
            teacher.q_status = teacher.QSTATUS_SOLVED
            ret = {'question': {}, 'answer': {}}
            if sdict['qtype'] == 0:
                ret['question']['name'] = "...."
                ret['answer']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['question']['music'] = r"\score{" \
                        r" { %s %s }" \
                        r"\layout { "\
                        r"  ragged-last = ##t " \
                        r"  \context { \Staff " \
                        r'\remove "Time_signature_engraver" } }' \
                        r"}" % (
                    teacher.m_tonika.get_octave_notename(),
                    (teacher.m_tonika + mpd.Interval.new_from_int(teacher.m_interval)).get_octave_notename())
                ret['answer']['music'] = ret['question']['music']
                yield ret
            else:
                assert sdict['qtype'] == 1
                ret['question']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['answer']['name'] = mpd.Interval.new_from_int(abs(teacher.m_interval)).get_name()
                ret['question']['music'] = r"\score{ { %s s4 s4} "\
                       r"\layout{ "\
                       r"  ragged-last = ##t "\
                       r"  \context { \Staff "\
                       r'     \remove "Time_signature_engraver" } }'\
                       r"}" % teacher.m_tonika.get_octave_notename()
                ret['answer']['music'] = r"\score{ { %s %s } "\
                       r"\layout{ "\
                       r"  ragged-last = ##t "\
                       r"  \context { \Staff "\
                       r'     \remove "Time_signature_engraver" } }'\
                       r"}" % (
                    teacher.m_tonika.get_octave_notename(),
                    (teacher.m_tonika + teacher.m_interval).get_octave_notename())
                yield ret
    def load_learning_tree_as_frontpage_data(self, filename):
        """
        This function can read the old "learning tree" format datafile and
        convert it to a front page structure.
        Load into the .m_frontpage_data variable
        """
        data = {}
        data['collections'] = {}
        data['columns'] = [[]]

        d = eval(open(filename, "r").read())
        for k in d['menu']:
            data['columns'][0].append({'heading': k['name'].replace("_", ""), 'collection-ids': []})
            for k2 in k['children']:
                newkey = "%s-%s" % (k['name'].lower().replace(" ", "_"), k2['name'].lower().replace(" ", "_"))
                assert newkey not in data
                data['collections'][newkey] = k2
                data['columns'][0][-1]['collection-ids'].append(newkey)
                del k2['nametotranslate']
                k2['name'] = k2['name'].replace("_", "")

        lines = sum(len(x['collection-ids'])+1 for x in data['columns'][0])
        count = 0
        for idx, x in enumerate(data['columns'][0]):
            count += len(x['collection-ids']) + 1
            if count > lines / 2:
                if len(data['columns']) == 1:
                    data['columns'].append([])
                    for i in range(idx+1, len(data['columns'][0])):
                        data['columns'][1].append(data['columns'][0][idx+1])
                        del data['columns'][0][idx+1]
                    break
        self.m_frontpage_data = data
